# Copyright 2018, Intel Corporation
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#
#     * Neither the name of the copyright holder nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import utils
from reorderexceptions import InconsistentFileException


class BinaryOutputHandler:
    """
    Handle :class:`BinaryFile` objects.

    Creates and aggregates :class:`BinaryFile` objects for ease of use.
    Implements methods for batch handling of aggregated files.

    :ivar _files: A list of registered files, most recent last.
    :type _files: list
    """

    def __init__(self, checker):
        """
        Binary handler constructor.

        :param checker: consistency checker object
        :type checker: ConsistencyCheckerBase
        """
        self._files = []
        self._checker = checker

    def add_file(self, file, map_base, size):
        """
        Create and append a mapped file to :attr:`_files`.

        :param file: Full path of the mapped file to be added.
        :type file: str
        :param map_base: Base address of the mapped file.
        :type map_base: int
        :param size: Size of the file.
        :type size: int
        :return: None
        """
        self._files.append(BinaryFile(file, map_base, size, self._checker))

    def remove_file(self, file):
        """Remove file from :attr:`_files`.

        :param file: File to be removed.
        :type file: str
        :return: None
        """
        for bf in self._files:
            if bf.file_name is file:
                self._files.remove(bf)

    def do_store(self, store_op):
        """
        Perform a store to the given file.

        The file is chosen based on the address and size
        of the store.

        :param store_op: The store operation to be performed.
        :type store_op: Store
        :return: None
        :raises: Generic exception - to be precised later.
        """
        store_ok = False
        for bf in self._files:
            if utils.range_cmp(store_op, bf) == 0:
                bf.do_store(store_op)
                store_ok = True
        if not store_ok:
            raise OSError(
                          "No suitable file found for store {}"
                          .format(store_op))

    def do_revert(self, store_op):
        """
        Reverts a store made to a file.

        Performing a revert on a store that has not been made
        previously yields undefined behavior.

        :param store_op: The store to be reverted.
        :type store_op: Store
        :return: None
        :raises: Generic exception - to be precised later.
        """
        revert_ok = False
        for bf in self._files:
            if utils.range_cmp(store_op, bf) == 0:
                bf.do_revert(store_op)
                revert_ok = True
        if not revert_ok:
            raise OSError(
                          "No suitable file found for store {}"
                          .format(store_op))

    def check_consistency(self):
        """
        Checks consistency of each registered file.

        :return: None
        :raises: Generic exception - to be precised later.
        """
        for bf in self._files:
            if not bf.check_consistency():
                raise InconsistentFileException(
                          "File {} inconsistent".format(bf))


class BinaryFile(utils.Rangeable):
    """Binary file handler.

    It is a handler for binary file operations. Internally it
    uses mmap to write to and read from the file.

    :ivar _file_name: Full path of the mapped file.
    :type _file_name: str
    :ivar _map_base: Base address of the mapped file.
    :type _map_base: int
    :ivar _map_max: Max address of the mapped file.
    :type _map_max: int
    :ivar _file_map: Memory mapped from the file.
    :type _file_map: mmap.mmap
    :ivar _checker: consistency checker object
    :type _checker: ConsistencyCheckerBase
    """

    def __init__(self, file_name, map_base, size, checker):
        """
        Initializes the binary file handler.

        :param file_name: Full path of the mapped file to be added.
        :type file_name: str
        :param map_base: Base address of the mapped file.
        :type map_base: int
        :param size: Size of the file.
        :type size: int
        :param checker: consistency checker object
        :type checker: ConsistencyCheckerBase
        :return: None
        """
        self._file_name = file_name
        self._map_base = map_base
        self._map_max = map_base + size
        # TODO consider mmaping only necessary parts on demand
        self._file_map = utils.memory_map(file_name)
        self._checker = checker

    def __str__(self):
        return self._file_name

    def do_store(self, store_op):
        """
        Perform the store on the file.

        The store records the old value for reverting.

        :param store_op: The store to be performed.
        :type store_op: Store
        :return: None
        """
        base_off = store_op.get_base_address() - self._map_base
        max_off = store_op.get_max_address() - self._map_base
        # read and save old value
        store_op.old_value = bytes(self._file_map[base_off:max_off])
        # write out the new value
        self._file_map[base_off:max_off] = store_op.new_value
        self._file_map.flush(base_off & ~4095, 4096)

    def do_revert(self, store_op):
        """
        Reverts the store.

        Write back the old value recorded while doing the store.
        Reverting a store which has not been made previously has
        undefined behavior.

        :param store_op: The store to be reverted.
        :type store_op: Store
        :return: None
        """
        base_off = store_op.get_base_address() - self._map_base
        max_off = store_op.get_max_address() - self._map_base
        # write out the old value
        self._file_map[base_off:max_off] = store_op.old_value
        self._file_map.flush(base_off & ~4095, 4096)

    def check_consistency(self):
        """
        Check consistency of the file.

        :return: True if consistent, False otherwise.
        :rtype: bool
        """
        return self._checker.check_consistency(self._file_name) == 0

    def get_base_address(self):
        """
        Returns the base address of the file.

        Overrides from :class:`utils.Rangeable`.

        :return: The base address of the mapping passed to the constructor.
        :rtype: int
        """
        return self._map_base

    def get_max_address(self):
        """
        Get max address of the file mapping.

        Overrides from :class:`utils.Rangeable`.

        :return: The max address of the mapping.
        :rtype: int
        """
        return self._map_max



# Copyright 2018, Intel Corporation
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#
#     * Neither the name of the copyright holder nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


from itertools import combinations
from itertools import permutations
from itertools import islice
from itertools import chain
from random import sample
from functools import partial
from reorderexceptions import NotSupportedOperationException
import collections


class FullReorderEngine:
    def __init__(self):
        self.test_on_barrier = True
    """
    Realizes a full reordering of stores within a given list.
    Example:
        input: (a, b, c)
        output:
               ()
               ('a',)
               ('b',)
               ('c',)
               ('a', 'b')
               ('a', 'c')
               ('b', 'a')
               ('b', 'c')
               ('c', 'a')
               ('c', 'b')
               ('a', 'b', 'c')
               ('a', 'c', 'b')
               ('b', 'a', 'c')
               ('b', 'c', 'a')
               ('c', 'a', 'b')
               ('c', 'b', 'a')
    """
    def generate_sequence(self, store_list):
        """
        Generates all possible combinations of all possible lengths,
        based on the operations in the list.

        :param store_list: The list of stores to be reordered.
        :type store_list: list of :class:`memoryoperations.Store`
        :return: Yields all combinations of stores.
        :rtype: iterable
        """
        for length in range(0, len(store_list) + 1):
            for permutation in permutations(store_list, length):
                yield permutation


class AccumulativeReorderEngine:
    def __init__(self):
        self.test_on_barrier = True
    """
    Realizes an accumulative reorder of stores within a given list.
    Example:
        input: (a, b, c)
        output:
               ()
               ('a')
               ('a', 'b')
               ('a', 'b', 'c')
    """
    def generate_sequence(self, store_list):
        """
        Generates all accumulative lists,
        based on the operations in the store list.

        :param store_list: The list of stores to be reordered.
        :type store_list: list of :class:`memoryoperations.Store`
        :return: Yields all accumulative combinations of stores.
        :rtype: iterable
        """

        for i in range(0, len(store_list) + 1):
            out_list = [store_list[i] for i in range(0, i)]
            yield out_list


class AccumulativeReverseReorderEngine:
    def __init__(self):
        self.test_on_barrier = True
    """
    Realizes an accumulative reorder of stores
    within a given list in reverse order.
    Example:
        input: (a, b, c)
        output:
               ()
               ('c')
               ('c', 'b')
               ('c', 'b', 'a')
    """
    def generate_sequence(self, store_list):
        """
        Reverse all elements order and
        generates all accumulative lists.

        :param store_list: The list of stores to be reordered.
        :type store_list: list of :class:`memoryoperations.Store`
        :return: Yields all accumulative combinations of stores.
        :rtype: iterable
        """
        store_list = list(reversed(store_list))
        for i in range(len(store_list) + 1):
            yield [store_list[j] for j in range(i)]


class SlicePartialReorderEngine:
    """
    Generates a slice of the full reordering of stores within a given list.
    Example:
        input: (a, b, c), start = 2, stop = None, step = 2
        output:
               ('b')
               ('a', 'b')
               ('b', 'c')
    """
    def __init__(self, start, stop, step=1):
        """
        Initializes the generator with the provided parameters.

        :param start: Number of preceding elements to be skipped.
        :param stop: The element at which the slice is to stop.
        :param step: How many values are skipped between successive calls.
        """
        self._start = start
        self._stop = stop
        self._step = step
        self.test_on_barrier = True

    def generate_sequence(self, store_list):
        """
        This generator yields a slice of all possible combinations.

        The result may be a set of combinations of different lengths,
        depending on the slice parameters provided at object creation.

        :param store_list: The list of stores to be reordered.
        :type store_list: list of :class:`memoryoperations.Store`
        :return: Yields a slice of all combinations of stores.
        :rtype: iterable
        """
        for sl in islice(chain(*map(lambda x: combinations(store_list, x),
                         range(0, len(store_list) + 1))),
                         self._start, self._stop, self._step):
            yield sl


class FilterPartialReorderEngine:
    """
    Generates a filtered set of the combinations
    without duplication of stores within a given list.
    Example:
        input: (a, b, c), filter = filter_min_elem, kwarg1 = 2
        output:
               (a, b)
               (a, c)
               (b, c)
               (a, b, c)

        input: (a, b, c), filter = filter_max_elem, kwarg1 = 2
        output:
               ()
               (a)
               (b)
               (c)
               (a, b)
               (a, c)
               (b, c)

        input: (a, b, c), filter = filter_between_elem, kwarg1 = 2, kwarg2 = 2
        output:
               (a, b)
               (a, c)
               (b, c)
    """
    def __init__(self, func, **kwargs):
        """
        Initializes the generator with the provided parameters.

        :param func: The filter function.
        :param **kwargs: Arguments to the filter function.
        """
        self._filter = func
        self._filter_kwargs = kwargs
        self.test_on_barrier = True

    @staticmethod
    def filter_min_elem(store_list, **kwargs):
        """
        Filter stores list if number of element is less than kwarg1
        """
        if (len(store_list) < kwargs["kwarg1"]):
            return False
        return True

    @staticmethod
    def filter_max_elem(store_list, **kwargs):
        """
        Filter stores list if number of element is greater than kwarg1.
        """
        if (len(store_list) > kwargs["kwarg1"]):
            return False
        return True

    @staticmethod
    def filter_between_elem(store_list, **kwargs):
        """
        Filter stores list if number of element is
        greater or equal kwarg1 and less or equal kwarg2.
        """
        store_len = len(store_list)
        if (store_len >= kwargs["kwarg1"] and store_len <= kwargs["kwarg2"]):
            return True
        return False

    def generate_sequence(self, store_list):
        """
        This generator yields a filtered set of combinations.

        :param store_list: The list of stores to be reordered.
        :type store_list: list of :class:`memoryoperations.Store`
        :return: Yields a filtered set of combinations.
        :rtype: iterable
        """
        filter_fun = getattr(self, self._filter, None)
        for elem in filter(
                       partial(filter_fun, **self._filter_kwargs), chain(
                           *map(lambda x: combinations(store_list, x), range(
                               0, len(store_list) + 1)))):
            yield elem


class RandomPartialReorderEngine:
    """
    Generates a random sequence of combinations of stores.
    Example:
        input: (a, b, c), max_seq = 3
        output:
               ('b', 'c')
               ('b',)
               ('a', 'b', 'c')
    """
    def __init__(self, max_seq=3):
        """
        Initializes the generator with the provided parameters.

        :param max_seq: The number of combinations to be generated.
        """
        self.test_on_barrier = True
        self._max_seq = max_seq

    def generate_sequence(self, store_list):
        """
        This generator yields a random sequence of combinations.
        Number of combinations without replacement has to be limited to
        1000 because of exponential growth of elements.
        Example:
            for 10 element from 80 -> 1646492110120 combinations
            for 20 element from 80 -> 3.5353161422122E+18 combinations
            for 40 element from 80 -> 1.0750720873334E+23 combinations
        :param store_list: The list of stores to be reordered.
        :type store_list: list of :class:`memoryoperations.Store`
        :return: Yields a random sequence of combinations.
        :rtype: iterable
        """
        population = list(chain(*map(
                          lambda x: islice(combinations(store_list, x), 1000),
                          range(0, len(store_list) + 1))))
        population_size = len(population)
        for elem in sample(population, self._max_seq if self._max_seq <=
                           population_size else population_size):
            yield elem


class NoReorderEngine:
    def __init__(self):
        self.test_on_barrier = True
    """
    A NULL reorder engine.
    Example:
        input: (a, b, c)
        output: (a, b, c)
    """
    def generate_sequence(self, store_list):
        """
        This generator does not modify the provided store list.

        :param store_list: The list of stores to be reordered.
        :type store_list: The list of :class:`memoryoperations.Store`
        :return: The unmodified list of stores.
        :rtype: iterable
        """
        return [store_list]


class NoCheckerEngine:
    def __init__(self):
        self.test_on_barrier = False
    """
    A NULL reorder engine.
    Example:
        input: (a, b, c)
        output: (a, b, c)
    """
    def generate_sequence(self, store_list):
        """
        This generator does not modify the provided store list
        and does not do the check.

        :param store_list: The list of stores to be reordered.
        :type store_list: The list of :class:`memoryoperations.Store`
        :return: The unmodified list of stores.
        :rtype: iterable
        """
        return [store_list]


def get_engine(engine):
    if engine in engines:
        reorder_engine = engines[engine]()
    else:
        raise NotSupportedOperationException(
                  "Not supported reorder engine: {}"
                  .format(engine))

    return reorder_engine


engines = collections.OrderedDict([
           ('NoReorderNoCheck', NoCheckerEngine),
           ('ReorderFull', FullReorderEngine),
           ('NoReorderDoCheck', NoReorderEngine),
           ('ReorderAccumulative', AccumulativeReorderEngine),
           ('ReorderReverseAccumulative', AccumulativeReverseReorderEngine),
           ('ReorderPartial', RandomPartialReorderEngine)])



# Copyright 2018, Intel Corporation
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#
#     * Neither the name of the copyright holder nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import logging

log_levels = ["debug", "info", "warning", "error", "critical"]


class LoggingBase:
    def debug(self, text):
        pass

    def info(self, text):
        pass

    def warning(self, text):
        pass

    def error(self, text):
        pass

    def critical(self, text):
        pass


class DefaultFileLogger(LoggingBase):
    def __init__(self, name="pmreorder", **kwargs):
        logging.basicConfig(**kwargs)
        self.__logger = logging.getLogger(name)

    def debug(self, text):
        self.__logger.debug(text)

    def info(self, text):
        self.__logger.info(text)

    def warning(self, text):
        self.__logger.warning(text)

    def error(self, text):
        self.__logger.error(text)

    def critical(self, text):
        self.__logger.critical(text)


class DefaultPrintLogger(LoggingBase):

    def debug(self, text):
        print("DEBUG:", text)

    def info(self, text):
        print("INFO:", text)

    def warning(self, text):
        print("WARNING:", text)

    def error(self, text):
        print("ERROR:", text)

    def critical(self, text):
        print("CRITICAL:", text)


def get_logger(log_output, log_level=None):
    logger = None
    # check if log_level is valid
    log_level = "warning" if log_level is None else log_level
    numeric_level = getattr(logging, log_level.upper())
    if not isinstance(numeric_level, int):
        raise ValueError('Invalid log level: %s'.format(log_level.upper()))

    if log_output is None:
        logger = DefaultPrintLogger()
    else:
        logger = DefaultFileLogger(filename=log_output, level=numeric_level)
    return logger



# Copyright 2018, Intel Corporation
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#
#     * Neither the name of the copyright holder nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from utils import Rangeable
from utils import range_cmp
from utils import StackTrace
from sys import byteorder


class BaseOperation:
    """
    Base class for all memory operations.
    """
    pass


class Fence(BaseOperation):
    """
    Describes a fence operation.

    The exact type of the memory barrier is not important,
    it is interpreted as an SFENCE or MFENCE.
    """
    class Factory:
        """
        Internal factory class to be used in dynamic object creation.
        """
        def create(self, values):
            """
            Factory object creation method.

            :param values: Ignored.
            :type values: str
            :return: New Fence object.
            :rtype: Fence
            """
            return Fence()


class Store(BaseOperation, Rangeable):
    """
    Describes a store operation.

    :ivar address: The virtual address at which to store the new value.
    :type address: int
    :ivar new_value: The new value to be written.
    :type new_value: bytearray
    :ivar size: The size of the store in bytes.
    :type size: int
    :ivar old_value: The old value read from the file.
    :type old_value: bytearray
    :ivar flushed: Indicates whether the store has been flushed.
    :type flushed: bool
    """
    def __init__(self, values):
        """
        Initializes the object based on the describing string.

        :param values: Pre-formatted string describing the store.
        :type values: str
        :return: None
        """
        params = values.split(";")
        # calculate the offset given the registered file mapping
        self.address = int(params[1], 16)
        self.size = int(params[3], 16)
        self.new_value = \
            int(params[2], 16).to_bytes(self.size, byteorder=byteorder)
        if len(params) > 4:
            self.trace = StackTrace(params[4:])
        else:
            self.trace = StackTrace(["No trace available", ])
        self.old_value = None
        self.flushed = False

    def __str__(self):
        return "addr: " + hex(self.address) + " size " + \
            str(self.size) + " value " + str(self.new_value)

    def get_base_address(self):
        """
        Override from :class:`utils.Rangeable`.

        :return: Virtual address of the store.
        :rtype: int
        """
        return self.address

    def get_max_address(self):
        """
        Override from :class:`utils.Rangeable`.

        :return: Virtual address of the first byte after the store.
        :rtype: int
        """
        return self.address + self.size

    class Factory():
        """
        Internal factory class to be used in dynamic object creation.
        """
        def create(self, values):
            """
            Factory object creation method.

            :param values: Pre-formatted string describing the store.
            :type values: str
            :return: New Store object.
            :rtype: Store
            """
            return Store(values)


class FlushBase(BaseOperation, Rangeable):
    """
    Base class for flush operations.
    """
    def is_in_flush(self, store_op):
        """
        Check if a given store is within the flush.

        :param store_op: Store operation to check.
        :return: True if store is in flush, false otherwise.
        :rtype: bool
        """
        raise NotImplementedError


class Flush(FlushBase):
    """
    Describes a flush operation.

    Examples of flush instructions are CLFLUSH, CLFLUSHOPT or CLWB.

    :ivar _address: Virtual address of the flush.
    :type _address: int
    :ivar _size: The size of the flush in bytes (should be cache line aligned).
    :type _size: int
    """
    def __init__(self, values):
        """
        Initializes the object based on the describing string.

        :param values: Pre-formatted string describing the flush.
        :type values: str
        :return: None
        """
        params = values.split(";")
        self._address = int(params[1], 16)
        self._size = int(params[2], 16)

    def is_in_flush(self, store_op):
        """
        Override from :class:`FlushBase`.

        :param store_op: Store operation to check.
        :return: True if store is in flush, false otherwise.
        :rtype: bool
        """
        if range_cmp(store_op, self) == 0:
            return True
        else:
            return False

    def get_base_address(self):
        """
        Override from :class:`utils.Rangeable`.

        :return: Virtual address of the flush.
        :rtype: int
        """
        return self._address

    def get_max_address(self):
        """
        Override from :class:`utils.Rangeable`.

        :return: Virtual address of the first byte after the flush.
        :rtype: int
        """
        return self._address + self._size

    class Factory:
        """
        Internal factory class to be used in dynamic object creation.
        """
        def create(self, values):
            """
            Factory object creation method.

            :param values: Pre-formatted string describing the flush.
            :type values: str
            :return: New Flush object.
            :rtype: Flush
            """
            return Flush(values)


class ReorderBase(BaseOperation):
    """
    Base class for all reorder type classes.
    """
    pass


class NoReorderDoCheck(ReorderBase):
    """
    Describes the type of reordering engine to be used.

    This marker class triggers writing the whole sequence of stores
    between barriers.
    """
    class Factory:
        """
        Internal factory class to be used in dynamic object creation.
        """
        def create(self, values):
            """
            Factory object creation method.

            :param values: Ignored.
            :type values: str
            :return: New NoReorderDoCheck object.
            :rtype: NoReorderDoCheck
            """
            return NoReorderDoCheck()


class ReorderFull(ReorderBase):
    """
    Describes the type of reordering engine to be used.

    This marker class triggers writing all possible sequences of stores
    between barriers.
    """
    class Factory:
        """
        Internal factory class to be used in dynamic object creation.
        """
        def create(self, values):
            """
            Factory object creation method.

            :param values: Ignored.
            :type values: str
            :return: New ReorderFull object.
            :rtype: ReorderFull
            """
            return ReorderFull()


class ReorderAccumulative(ReorderBase):
    """
    Describes the type of reordering engine to be used.

    This marker class triggers writing all
    possible accumulative sequences of stores
    between barriers.
    """
    class Factory:
        """
        Internal factory class to be used in dynamic object creation.
        """
        def create(self, values):
            """
            Factory object creation method.

            :param values: Ignored.
            :type values: str
            :return: New ReorderAccumulative object.
            :rtype: ReorderAccumulative
            """
            return ReorderAccumulative()


class ReorderReverseAccumulative(ReorderBase):
    """
    Describes the type of reordering engine to be used.

    This marker class triggers writing all
    possible reverted accumulative sequences of stores
    between barriers.
    """
    class Factory:
        """
        Internal factory class to be used in dynamic object creation.
        """
        def create(self, values):
            """
            Factory object creation method.

            :param values: Ignored.
            :type values: str
            :return: New ReorderReverseAccumulative object.
            :rtype: ReorderReverseAccumulative
            """
            return ReorderReverseAccumulative()


class NoReorderNoCheck(ReorderBase):
    """
    Describes the type of reordering engine to be used.

    This marker class triggers writing the whole sequence of stores
    between barriers. It additionally marks that no consistency checking
    is to be made.
    """
    class Factory:
        """
        Internal factory class to be used in dynamic object creation.
        """
        def create(self, values):
            """
            Factory object creation method.

            :param values: Ignored.
            :type values: str
            :return: New NoReorderNoCheck object.
            :rtype: NoReorderNoCheck
            """
            return NoReorderNoCheck()


class ReorderDefault(ReorderBase):
    """
    Describes the default reordering engine to be used.

    This marker class triggers default reordering.
    """
    class Factory:
        """
        Internal factory class to be used in dynamic object creation.
        """
        def create(self, values):
            """
            Factory object creation method.

            :param values: Ignored.
            :type values: str
            :return: ReorderDefault object.
            :rtype: ReorderDefault
            """
            return ReorderDefault()


class ReorderPartial(ReorderBase):
    """
    Describes the type of reordering engine to be used.

    This marker class triggers writing a subset of all possible
    sequences of stores between barriers.

    The type of partial reordering is chosen at runtime. Not yet
    implemented.
    """
    class Factory:
        """
        Internal factory class to be used in dynamic object creation.
        """
        def create(self, values):
            """
            Factory object creation method.

            :param values: Ignored.
            :type values: str
            :return: New ReorderPartial object.
            :rtype: ReorderPartial
            """
            return ReorderPartial()


class Register_file(BaseOperation):
    """
    Describes the file to be mapped into processes address space.

    :ivar name: The full name of the file.
    :type name: str
    :ivar address: The base address where the file was mapped.
    :type address: int
    :ivar size: The size of the mapping.
    :type size: int
    :ivar offset: The start offset of the mapping within the file.
    :type offset: int
    """
    def __init__(self, values):
        """
        Initializes the object based on the describing string.

        :param values: Pre-formatted string describing the flush.
        :type values: str
        :return: None
        """
        params = values.split(";")
        self.name = params[1]
        self.address = int(params[2], 16)
        self.size = int(params[3], 16)
        self.offset = int(params[4], 16)

    class Factory():
        """
        Internal factory class to be used in dynamic object creation.
        """
        def create(self, values):
            """
            Factory object creation method.

            :param values: Pre-formatted string
                           describing the file registration.
            :type values: str
            :return: New Register_file object.
            :rtype: Register_file
            """
            return Register_file(values)



# Copyright 2018, Intel Corporation
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#
#     * Neither the name of the copyright holder nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import os
import json


class MarkerParser:
    """
    Parse marker config file and command line arg provided by user
    via -x parameter.
    """
    def marker_file_parser(self, macros):
        """
        Parse markers passed by file.
        They should be in json format:
        { "MARKER_NAME"="ENGINE_TYPE" } and separated by commas.
        """
        markers = {}
        try:
            with open(macros) as config_file:
                markers = json.load(config_file)
        except json.decoder.JSONDecodeError:
            print("Invalid config macros file format: ", macros,
                  "Use: {\"MARKER_NAME1\"=\"ENGINE_TYPE1\","
                  "\"MARKER_NAME2\"=\"ENGINE_TYPE2\"}")

        return markers

    def marker_cli_parser(self, macros):
        """
        Parse markers passed by cli.
        They should be in specific format:
        MARKER_NAME=ENGINE_TYPE and separated by commas.
        """
        try:
            markers_array = macros.split(",")
            return dict(pair.split('=') for pair in markers_array)
        except ValueError:
            print("Invalid extended macros format: ", macros,
                  "Use: MARKER_NAME1=ENGINE_TYPE1,MARKER_NAME2=ENGINE_TYPE2")

    def get_markers(self, markerset):
        """
        Parse markers based on their format.
        """
        if markerset is not None:
            if os.path.exists(markerset):
                return self.marker_file_parser(markerset)
            else:
                return self.marker_cli_parser(markerset)
